from auto_genetic.population_initializer.chromosomes import Chromosome
from typing import List


class PopulationInitializer:
    """
    Class used to initialize the population, made of chromosomes
    Attributes
    ----------
    pop_size: int
              the number of chromosomes we want in our population
    algo_type: str
               can be "hyperparameter_tuning" or "feature_selection", it represents what is the purpose of the GA
    hyperparams_dict: List[dict, dict]
                      a list that contains 2 dictionaries, one with the hyperparameters' types and another with the
                      hyperparameters' values
    feature_num: int
                 the number of features from which to select


    Methods
    -------
        self.initialize_population(): method that uses the Chromosome class to generate n (= pop_size) Chromosomes
    """
    def __init__(self, pop_size: int, algo_type: str, hyperparams_dict: list = None, feature_num: int = None) -> None:
        self.population = []
        self.algo_type = algo_type
        self.hyperparams_dict = hyperparams_dict
        self.feature_num = feature_num
        self.pop_size = pop_size

    def initialize_population(self) -> List[Chromosome]:
        """
        method that uses the Chromosome class to generate n (= pop_size) Chromosomes
        Returns
        -------
        List[Chromosome]
        a list that contains all the chromosomes for the first generation

        """
        for _ in range(self.pop_size):
            chromosome = Chromosome(self.algo_type, hyperparams_dict=self.hyperparams_dict,
                                    feature_num=self.feature_num).initialize()

            self.population.append(chromosome)

        return self.population



import numpy as np
import random
from auto_genetic.population_initializer.chromosomes import Chromosome
from copy import deepcopy


class Mutation:
    """
    Class that performs the crossover on the parents to get the children
    Attributes
    ----------
    mutation_method: str
                      the methods that we can use to perform the mutation, it can be "bit_flip", "random",
                      "swap", "scramble" or "inversion"
    prob_mutation: float
                    the probability that the mutation is performed
    hyperparams_types: dict
                       a dictionary that has the names of the hyperparameters as keys and the hyperparameter types as
                       values, this can either be "continuous" or "categorical"
    Methods
    -------
    Private:
        self._bit_flip_mutation(chromosome): it performs the bit flip mutation on the selected chromosome
        self._random_resetting_mutation(chromosome): it performs the random resetting mutation on the selected
                                                     chromosome
        self._swap_mutation(chromosome): it performs the swap mutation on the selected chromosome
        self._scramble_mutation(chromosome): it performs the scramble mutation on the selected chromosome
        self._inversion_mutation(chromosome): it performs the inversion mutation on the selected chromosome
        self._random_categorical(chromosome): it chooses a new random value for the value that needs to be mutated
                                              amongst the possible one for the specified categorical parameter
        self._random_continuous(chromosome): it chooses a new random value for the value that needs to be mutated
                                             amongst the possible one for the specified continuous parameter

    Public:
        self.perform_mutation(chromosome): it performs the mutation by feeding the chromosome to the
                                            previously specified mutation method
    """
    def __init__(self, mutation_method: str, hyperparams_types: dict, prob_mutation: float) -> None:
        self.mutation_method = mutation_method
        self.prob_mutation = prob_mutation
        self.hyperparams_types = hyperparams_types
        self.hyperparams_map = None

    def perform_mutation(self, chromosome: Chromosome) -> Chromosome:
        """
        it performs the mutation by feeding the chromosome to the previously specified mutation method
        Parameters
        ----------
        chromosome: Chromosome
                    the chromosome that contains the sequence over which to apply the mutation

        Returns
        -------
        Chromosome
        the previous chromosome but with a mutated sequence
        """
        self.hyperparams_map = deepcopy(chromosome.hyperparams_map)
        if self.mutation_method == "bit_flip":
            mutated_gene = self._bit_flip_mutation(chromosome)

        elif self.mutation_method == "random":
            mutated_gene = self._random_resetting_mutation(chromosome)

        elif self.mutation_method == "swap":
            mutated_gene = self._swap_mutation(chromosome)

        elif self.mutation_method == "scramble":
            mutated_gene = self._scramble_mutation(chromosome)

        elif self.mutation_method == "inversion":
            mutated_gene = self._inversion_mutation(chromosome)

        else:
            raise ValueError("the mutation method specified is not implemented")

        mutated_gene.fitness = None

        return mutated_gene

    def _bit_flip_mutation(self, chromosome: Chromosome) -> np.array:
        """
        it performs the bit flip mutation on the selected chromosome
        Parameters
        ----------
        chromosome: Chromosome
                    the chromosome that contains the sequence over which to apply the bit flip mutation

        Returns
        -------
        Chromosome
        the previous chromosome but with a mutated sequence
        """
        to_be_mutated = np.random.rand(*chromosome.sequence.shape) <= self.prob_mutation
        mutation_index = np.argwhere(to_be_mutated)
        mutated_gene = deepcopy(chromosome)

        for i in mutation_index:
            if mutated_gene.sequence[i] == 0:
                mutated_gene.sequence[i] = 1
            else:
                mutated_gene.sequence[i] = 1

        return mutated_gene

    def _random_resetting_mutation(self, chromosome: Chromosome) -> Chromosome:
        """
        it performs the random resetting mutation on the selected chromosome
        Parameters
        ----------
        chromosome: Chromosome
                    the chromosome that contains the sequence over which to apply the random resetting mutation

        Returns
        -------
        Chromosome
        the previous chromosome but with a mutated sequence
        """
        to_be_mutated = np.random.rand(*chromosome.sequence.shape) <= self.prob_mutation
        mutation_index = np.argwhere(to_be_mutated).flatten()
        mutated_gene = deepcopy(chromosome)

        for i in mutation_index:
            param = list(self.hyperparams_map.keys())[i]
            if self.hyperparams_types[param] == "continuous":
                choice = self._random_continuous(param)

            else:
                choice = self._random_categorical(param)

            mutated_gene.sequence[i] = choice

        return mutated_gene

    def _swap_mutation(self, chromosome: Chromosome) -> Chromosome:
        """
        it performs the swap mutation on the selected chromosome
        Parameters
        ----------
        chromosome: Chromosome
                    the chromosome that contains the sequence over which to apply the swap mutation

        Returns
        -------
        Chromosome
        the previous chromosome but with a mutated sequence
        """
        to_be_swapped = np.random.rand(*chromosome.sequence.shape) <= self.prob_mutation
        swapped_index = np.argwhere(to_be_swapped)
        if swapped_index % 2 == 1:
            swapped_index = np.random.choice(swapped_index, size=len(swapped_index)-1, replace=False)

        mutated_gene = deepcopy(chromosome)

        for i in range(0, len(swapped_index), 2):
            mutated_gene.sequence[swapped_index[i]], mutated_gene.sequence[swapped_index[i + 1]] = \
                mutated_gene.sequence[swapped_index[i]], mutated_gene.sequence[swapped_index[i + 1]]

        return mutated_gene

    def _scramble_mutation(self, chromosome: Chromosome) -> Chromosome:
        """
        it performs the scramble mutation on the selected chromosome
        Parameters
        ----------
        chromosome: Chromosome
                    the chromosome that contains the sequence over which to apply the scramble mutation

        Returns
        -------
        Chromosome
        the previous chromosome but with a mutated sequence
        """
        prob = np.random.rand()
        mutated_child = deepcopy(chromosome)
        if prob <= self.prob_mutation:
            index_1 = random.randint(0, len(chromosome.sequence))
            index_2 = random.randint(index_1, len(chromosome.sequence))

            mutated_child.sequence = np.hstack([mutated_child.sequence[:index_1],
                                               random.shuffle(mutated_child.sequence[index_1:index_2]),
                                               mutated_child.sequence[index_2:]])

        return mutated_child

    def _inversion_mutation(self, chromosome: Chromosome) -> Chromosome:
        """
        it performs the inversion mutation on the selected chromosome
        Parameters
        ----------
        chromosome: Chromosome
                    the chromosome that contains the sequence over which to apply the inversion mutation

        Returns
        -------
        Chromosome
        the previous chromosome but with a mutated sequence
        """
        prob = np.random.rand()
        mutated_child = deepcopy(chromosome)
        if prob <= self.prob_mutation:
            index_1 = random.randint(0, len(chromosome.sequence))
            index_2 = random.randint(index_1, len(chromosome.sequence))

            mutated_child.sequence = np.hstack([mutated_child.sequence[:index_1],
                                               mutated_child.sequence[index_1:index_2][::-1],
                                               mutated_child.sequence[index_2:]])

        return mutated_child

    def _random_categorical(self, param: str) -> int:
        """
        it chooses a new random value for the value that needs to be mutated amongst the possible one for the
        specified categorical parameter
        Parameters
        ----------
        param: str
               the name of the parameter that we will choose a random value for

        Returns
        -------
        int
        the new value for the parameter chosen
        """
        values = self.hyperparams_map[param][0]
        choice = np.random.choice(values)

        return choice

    def _random_continuous(self, param: str) -> float:
        """
        it chooses a new random value for the value that needs to be mutated amongst the possible one for the
        specified continuous parameter
        Parameters
        ----------
        param: str
               the name of the parameter that we will choose a random value for

        Returns
        -------
        float
        the new value for the parameter chosen
        """
        values = self.hyperparams_map[param]
        choice = random.uniform(values[0], values[1])

        return choice



import random
from auto_genetic.population_initializer.chromosomes import Chromosome


class Darwinism:
    """
    Class that performs the reproduction over the whole population
    Attributes
    ----------
    population: list
                the list that contains all the Chromosomes that make up the population
    reproduction_rate: flaat
                       the rate at which the population reproduces itself, the number of children is equal to number
                       of chromosome in the generation * reproduction_rate
    Methods
    -------
    self.discard_worst_performers(): it eliminates the chromosomes with the worst fitness from the generation
    """
    def __init__(self, population: list, reproduction_rate: float) -> None:
        self.population = population
        self.reproduction_rate = reproduction_rate
        self.best_chromosome = None

    def discard_worst_performers(self) -> tuple[list, Chromosome]:
        """
        it eliminates the chromosomes with the worst fitness from the generation
        Returns
        -------
        tuple[list, Chromosome]
        a tuple that contains the population without the worst performers and the best_chromosome in the population
        """
        keep_number = int(len(self.population) * (1 - self.reproduction_rate))

        self.population.sort(key=lambda chromosome: chromosome.fitness, reverse=True)
        self.population = self.population[:keep_number]
        self.best_chromosome = self.population[0]
        random.shuffle(self.population)

        return self.population, self.best_chromosome



from auto_genetic.population_initializer.chromosomes import Chromosome
from auto_genetic.evaluation.fitnessfuncsetter import FitnessFuncSetter
from auto_genetic.evaluation.feature_selector import FeatureSelector
from tqdm import tqdm
import pandas as pd


class Evaluator:
    """
    Class for the evaluation of chromosomes
    Attributes
    ----------
    program: class
            the user defined program that has the run() and predict() methods
    population: list
                the list that contains all the Chromosomes that make up the population
    evaluation_method: str
                       the string that defines the evaluation method that we want to use inside the evaluator,
                       if it's 'custom' the user needs to define the custom_fitness_function
    target: str
            the name of the column that is used as target for the evaluation
    algo_type: str
               can be "hyperparameter_tuning" or "feature_selection", it represents what is the purpose of the GA
    train_data: pd.DataFrame
                a pandas dataframe that contains the training data, including the target
    test_data: pd.DataFrame
               a pandas dataframe that contains the training data, possibly also the target (depends on the
               fitness function)
    custom_fitness_function: class
                             if evaluation_method is custom here we have to have a custom evaluation function

    Methods
    -------
    Private:
        self._set_program_hyperparams(chromosome): it sets the hyperparameters for the program using a specific
                                                   chromosome
        self._filter_data(chromosome): it filters the data for the dataframe for a certain chromosome
    Public:
        self.evaluate_chromosome(chromosome): it computes the fitness for a certain chromosome
        self.evaluate_generation(): it computes the fitness for the whole population
    """

    def __init__(self, program, population: list, evaluation_method: str, target: str, algo_type: str,
                 train_data: pd.DataFrame, test_data: pd.DataFrame, custom_fitness_function=None) -> None:

        self.program = program
        self.population = population
        self.evaluation_method = evaluation_method
        self.train_data = train_data
        self.test_data = test_data
        self.target = target
        self.custom_fitness_function = custom_fitness_function
        self.algo_type = algo_type

    def _set_program_hyperparams(self, chromosome: Chromosome) -> None:
        """
        it sets the hyperparameters of a certain chromosome
        Parameters
        ----------
        chromosome: Chromosome
                    A chromosome that contains the hyperparameters to feed to our program

        """
        if chromosome.hyperparams is None:
            raise ValueError("There is an issue with the hyperparameters of this chromosome")

        self.program.set_program_hyperparams(chromosome.hyperparams)

    def _filter_data(self, chromosome: Chromosome) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        it uses the sequence of the chromosome to select which columns will be selected from the dataframe
        Parameters
        ----------
        chromosome: Chromosome
                    A chromosome that contains a sequence of 0s and 1s to activate/dis-activate the columns

        """
        selector_train = FeatureSelector(data=self.train_data, target=self.target)
        selector_test = FeatureSelector(data=self.test_data, target=self.target)

        train_data, test_data = selector_train.feature_select(chromosome), selector_test.feature_select(
            chromosome)

        return train_data, test_data

    def evaluate_chromosome(self, chromosome: Chromosome) -> None:
        """
        Function to evaluate the fitness of a chromosome by running the program using the predefined fitness function
        Parameters
        ----------
        chromosome: Chromosome
                    contains the information needed for the evaluation and will receive a fitness attribute

        """
        if self.algo_type == "feature_selection":
            train_data, test_data = self._filter_data(chromosome)
            func_setter = FitnessFuncSetter(self.evaluation_method, self.program, train_data, test_data,
                                            self.target)
        elif self.algo_type == "hyperparameter_tuning":
            self._set_program_hyperparams(chromosome)
            func_setter = FitnessFuncSetter(self.evaluation_method, self.program, self.train_data, self.test_data,
                                            self.target)

        else:
            raise ValueError("Please specify a valid algo_type")

        fitness_function = func_setter.set_fitness_func(
            custom_fitness_function=self.custom_fitness_function).get_fitness_func()
        chromosome.set_fitness_function(fitness_function)

        chromosome.calculate_fitness()

    def evaluate_generation(self):
        """
        Method to evaluate the full generation by using self.evaluate_chromosome for every member of the generation
        """
        for chromosome in tqdm(self.population, miniters=10):
            self.evaluate_chromosome(chromosome)



from __future__ import annotations

import numpy as np
import random
from typing import List


class HyperParamsSetter:
    """
    Class used to set the hyperparameters for a specific chromosome
    Attributes
    ----------
    algo_type: str
               can be "hyperparameter_tuning" or "feature_selection", it represents what is the purpose of the GA
    hyperparams_dict: List[dict, dict]
                      a list that contains 2 dictionaries, one with the hyperparameters' types and another with the
                      hyperparameters' values
    sequence: list
              the list of the values of the chromosome's gene


    Methods
    -------
    Private:
        self._set_continuous(): it chooses the value in the sequence for a continuous hyperparameter
        self._set_categorical(): it chooses the value in the sequence for a categorical hyperparameter
    Public:
        self.get_hyperparameters(): it extracts the hyperparameters from the hyperparameter dict
        self.convert_hyperparams_values(): it sets the sequence for the chromosome, given the hyperparameters
        self.get_program_hyperparams(): it extracts the specific hyperparameters from a chromosome,
                                        that were encoded according to the hyperparameter map
    """
    def __init__(self, hyperparams_dict: List[dict, dict], algo_type: str) -> None:
        self.hyperparams_dict = hyperparams_dict
        self.algo_type = algo_type
        self.hyperparams_values = None
        self.hyperparams_names = None
        self.hyperparams_types = None
        self.hyperparams_map = None
        self.sequence = None

    def get_hyperparameters(self) -> HyperParamsSetter:
        """
        it extracts the hyperparameters from the hyperparameter dict
        Returns
        -------
        HyperParamsSetter
        returns the same class with the more attributes set
        """
        if self.algo_type == "hyperparameter_tuning":
            self.hyperparams_names = self.hyperparams_dict[0].keys()
            self.hyperparams_values = self.hyperparams_dict[1]
            self.hyperparams_types = self.hyperparams_dict[0]

        else:
            raise ValueError("If the algo_type is not hyperparameter_tuning, you should not need this method")
        return self

    def convert_hyperparams_values(self) -> np.array:
        """
        it sets the sequence for the chromosome, given the hyperparameters
        Returns
        -------
        np.array
        the sequence for the chromosome, encoded with the hyperparams map

        """
        self.hyperparams_map = {}
        sequence = []
        for param, values in self.hyperparams_values.items():
            if self.hyperparams_types[param] == "categorical":
                choice = self._set_categorical(param, values)
                sequence.append(choice)
            elif self.hyperparams_types[param] == "continuous":
                choice = self._set_continuous(param, values)
                sequence.append(choice)

            else:
                raise ValueError(f"The parameter: {param}, is neither categorical nor continuous, check the "
                                 f"hyperparameter types and restart")
        self.sequence = sequence

        return np.array(sequence)

    def get_program_hyperparams(self) -> tuple:
        """
        it extracts the specific hyperparameters from a chromosome, that were encoded according to the
        hyperparameter map
        Returns
        -------
        tuple
        A tuple containing 2 dictionaries, the specific hyperparams for the chromosome and the hyperparameter map

        """
        i = 0
        hyperparams = {}
        for name in self.hyperparams_names:
            index = self.sequence[i]
            if self.hyperparams_types[name] == "categorical":
                parameter = self.hyperparams_map[name][1][index]

            else:
                parameter = index

            hyperparams[name] = parameter
            i += 1
        return hyperparams, self.hyperparams_map

    def _set_continuous(self, param: str, values: list) -> float:
        """
        it chooses the value in the sequence for a continuous hyperparameter
        Parameters
        ----------
        param: str
               a string that identifies the parameter we are currently considering
        values: list
                the range between which our hyperparameter can be set

        Returns
        -------
        flaat
        the value that the parameter will take in the program
        """

        self.hyperparams_map[param] = values
        choice = random.uniform(values[0], values[1])

        return choice

    def _set_categorical(self, param: str, values: list) -> int:
        """
        it chooses the value in the sequence for a categorical hyperparameter
        Parameters
        ----------
        param: str
               a string that identifies the parameter we are currently considering
        values: list
                the list of values that our hyperparameter can take

        Returns
        -------
        int
        the encoded value that the parameter will take in the program, corresponding to its position in the
        hyperparameters map
        """
        choices = list(range(len(values)))
        self.hyperparams_map[param] = [choices, values]
        choice = np.random.choice(choices)

        return choice



import numpy as np
from auto_genetic.reproduction.crossover import CrossOver
from auto_genetic.reproduction.mutation import Mutation
from auto_genetic.reproduction.translation import Translation
from auto_genetic.reproduction.selection import Selection
from auto_genetic.population_initializer.chromosomes import Chromosome


class Reproduction:
    """
    Class that performs the reproduction over the whole population
    Attributes
    ----------
    population: list
                the list that contains all the Chromosomes that make up the population
    crossover_method: str
                      the methods that we can use to perform the crossover, it can be "single_point_split",
                      "two_point_split" or "uniform"
    prob_crossover: float
                    the probability that the crossover is performed, by default it's equal to 1, so the crossover
                    will always be applied
    hyperparams_types: dict
                       a dictionary that has the names of the hyperparameters as keys and the hyperparameter types as
                       values, this can either be "continuous" or "categorical"
    mutation_method: str
                      the methods that we can use to perform the mutation, it can be "bit_flip", "random",
                      "swap", "scramble" or "inversion"
    prob_mutation: float
                    the probability that the mutation is performed
    prob_translation: float
                      the probability that the translation is performed
    selection_method: str
                      the method that we can use to perform the selection of the parents, it can be "roulette_wheel",
                      "stochastic_universal_sampling", "tournament", "rank" or "random"
    reproduction_rate: flaat
                       the rate at which the population reproduces itself, the number of children is equal to number
                       of chromosome in the generation * reproduction_rate
    tournament_size: int
                     the size of the tournament that we use for the "tournament" selection
    Methods
    -------
    self.crossover(parent_1, parent_2): it uses the Crossover class on 2 Chromosomes and performs the crossover
    self.mutation(child): it uses the Mutation class on a Chromosome and performs the mutation
    self.translation(child): it uses the Translation class on a Chromosome and performs the translation
    self.selection(): it uses the Selection class on the population and it selects the parents for the next generation
    self.reproduction(): it puts together all the steps necessary for the reproduction and generates the children that
                         will be added to the population
    """
    def __init__(self, population: np.array = None, crossover_method: str = "single_point_split",
                 prob_crossover: float = 1, hyperparams_types: dict = None, mutation_method: str = "bit_flip",
                 prob_mutation: float = 0.3,
                 prob_translation: float = 0.1, selection_method: str = "roulette_wheel",
                 reproduction_rate: float = 0.2, tournament_size: int = 4) -> None:
        self.prob_crossover = prob_crossover
        self.crossover_method = crossover_method
        self.prob_mutation = prob_mutation
        self.prob_translation = prob_translation
        self.reproduction_rate = reproduction_rate
        self.hyperparams_types = hyperparams_types
        self.population = population
        self.selection_method = selection_method
        self.tournament_size = tournament_size
        self.mutation_method = mutation_method

    def crossover(self, parent_1: Chromosome, parent_2: Chromosome) -> tuple[Chromosome, Chromosome]:
        """
        it uses the Crossover class on 2 Chromosomes and performs the crossover
        Parameters
        ----------
        parent_1: Chromosome
                  one of the chromosomes selected to become parents by the selection algorithm
        parent_2: Chromosome
                  one of the chromosomes selected to become parents by the selection algorithm

        Returns
        -------
        tuple[Chromosome, Chromosome]
        a tuple containing the 2 children, obtained by performing crossover
        """
        child1, child2 = CrossOver(self.crossover_method).perform_crossover(parent_1=parent_1, parent_2=parent_2)

        return child1, child2

    def mutation(self, child: Chromosome) -> Chromosome:
        """
        it uses the Mutation class on a Chromosome and performs the mutation
        Parameters
        ----------
        child: Chromosome
                the chromosome that contains the sequence over which to apply the mutation

        Returns
        -------
        Chromosome
        the previous chromosome but with a mutated sequence
        """
        mutated_child = Mutation(self.mutation_method, self.hyperparams_types,
                                 self.prob_mutation).perform_mutation(child)

        return mutated_child

    def translation(self, child: Chromosome) -> Chromosome:
        """
        it uses the Translation class on a Chromosome and performs the translation
        Parameters
        ----------
        child: Chromosome
               the chromosome that contains the sequence over which to apply the translation


        Returns
        -------
        Chromosome
        the previous chromosome but with a translated sequence
        """
        translated_child = Translation(self.prob_translation).perform_translation(child)

        return translated_child

    def selection(self) -> np.array:
        """
        it uses the Selection class on the population and it selects the parents for the next generation
        Returns
        -------
        np.array
        an array that contains the selected parents for the reproduction
        """
        parents = Selection(self.selection_method, self.population,
                            self.reproduction_rate).perform_selection(tournament_size=self.tournament_size)

        return parents

    def reproduction(self) -> list:
        """
        it puts together all the steps necessary for the reproduction and generates the children that
        will be added to the population
        Returns
        -------
        list
        the list of all the children that have been created during the reproduction
        """
        parents = self.selection()
        children = []

        for i in range(0, len(parents), 2):
            parent_1, parent_2 = parents[i], parents[i + 1]
            child_1, child_2 = self.crossover(parent_1, parent_2)
            child_1, child_2 = self.mutation(child_1), self.mutation(child_2)
            child_1, child_2 = self.translation(child_1), self.translation(child_2)
            children.append(child_1)
            children.append(child_2)

        return children



import numpy as np
import random
from auto_genetic.population_initializer.chromosomes import Chromosome
from copy import deepcopy


# TODO handle the fact that parent and children are now classes and not np.array
class CrossOver:
    """
    Class that performs the crossover on the parents to get the children
    Attributes
    ----------
    crossover_method: str
                      the methods that we can use to perform the crossover, it can be "single_point_split",
                      "two_point_split" or "uniform"
    prob_crossover: float
                    the probability that the crossover is performed, by default it's equal to 1, so the crossover
                    will always be applied
    Methods
    -------
    Private:
        self._single_split_crossover(parent_1, parent_2): it performs the single split crossover on the 2 parents to
                                                          get the children
        self._single_split_crossover(parent_1, parent_2): it performs the two split crossover on the 2 parents to
                                                          get the children
        self._uniform_crossover(parent_1, parent_2): it performs a uniform crossover on the 2 parents to get
                                                     the children
    Public:
        self.perform_crossover(parent_1, parent_2): it performs the crossover by feeding the parents to the
                                                    previously specified crossover method
    """
    def __init__(self, crossover_method: str, prob_crossover: float = 1.0) -> None:
        self.crossover_method = crossover_method
        self.prob_crossover = prob_crossover

    def perform_crossover(self, parent_1: Chromosome, parent_2: Chromosome) -> tuple[Chromosome, Chromosome]:
        """
        it performs the crossover by feeding the parents to the previously specified crossover method
        Parameters
        ----------
        parent_1: Chromosome
                  one of the chromosomes selected to become parents by the selection algorithm
        parent_2: Chromosome
                  one of the chromosomes selected to become parents by the selection algorithm

        Returns
        -------
        tuple[Chromosome, Chromosome]
        a tuple containing the 2 children, obtained by performing crossover
        """

        if self.crossover_method == "single_point_split":
            child_1, child_2 = self._single_split_crossover(parent_1, parent_2)

        elif self.crossover_method == "two_point_split":
            child_1, child_2 = self._two_split_crossover(parent_1, parent_2)

        elif self.crossover_method == "uniform":
            child_1, child_2 = self._uniform_crossover(parent_1, parent_2)
        else:
            raise ValueError("the crossover method selected has not been implemented")

        child_1.fitness = None
        child_2.fitness = None

        return child_1, child_2

    def _single_split_crossover(self, parent_1: Chromosome, parent_2: Chromosome) -> tuple[Chromosome, Chromosome]:
        """
        it performs the single split crossover on the 2 parents to get the children
        Parameters
        ----------
        parent_1: Chromosome
                  one of the chromosomes selected to become parents by the selection algorithm
        parent_2: Chromosome
                  one of the chromosomes selected to become parents by the selection algorithm

        Returns
        -------
        tuple[Chromosome, Chromosome]
        a tuple containing the 2 children, obtained by performing crossover
        """

        prob = np.random.rand()

        if prob <= self.prob_crossover:
            # generating the random number to perform crossover
            crossover_point = random.randint(0, len(parent_1.sequence))

            # interchanging the genes
            child1 = np.hstack([parent_1.sequence[: crossover_point], parent_2.sequence[crossover_point:]])
            child2 = np.hstack([parent_2.sequence[: crossover_point], parent_1.sequence[crossover_point:]])

            child_1 = deepcopy(parent_1)
            child_2 = deepcopy(parent_2)

            child_1.sequence = child1
            child_2.sequence = child2

            return child_1, child_2

        else:
            return deepcopy(parent_1), deepcopy(parent_2)

    def _two_split_crossover(self, parent_1: Chromosome, parent_2: Chromosome) -> tuple[Chromosome, Chromosome]:
        """
        it performs the two point split crossover on the 2 parents to get the children
        Parameters
        ----------
        parent_1: Chromosome
                  one of the chromosomes selected to become parents by the selection algorithm
        parent_2: Chromosome
                  one of the chromosomes selected to become parents by the selection algorithm

        Returns
        -------
        tuple[Chromosome, Chromosome]
        a tuple containing the 2 children, obtained by performing crossover
        """
        prob = np.random.rand()

        if prob <= self.prob_crossover:
            index_1 = random.randint(0, len(parent_1.sequence)//2)
            index_2 = random.randint(index_1, len(parent_1.sequence))

            child1 = np.hstack([parent_1.sequence[: index_1], parent_2.sequence[index_1: index_2],
                               parent_1.sequence[index_2:]])
            child2 = np.hstack([parent_2.sequence[: index_1], parent_1.sequence[index_1: index_2],
                               parent_2.sequence[index_2:]])

            child_1 = deepcopy(parent_1)
            child_2 = deepcopy(parent_2)

            child_1.sequence = child1
            child_2.sequence = child2

            return child_1, child_2

        else:
            return deepcopy(parent_1), deepcopy(parent_2)

    def _uniform_crossover(self, parent_1: Chromosome, parent_2: Chromosome) -> tuple[Chromosome, Chromosome]:
        """
        it performs the uniform crossover on the 2 parents to get the children
        Parameters
        ----------
        parent_1: Chromosome
                  one of the chromosomes selected to become parents by the selection algorithm
        parent_2: Chromosome
                  one of the chromosomes selected to become parents by the selection algorithm

        Returns
        -------
        tuple[Chromosome, Chromosome]
        a tuple containing the 2 children, obtained by performing crossover
        """
        prob = np.random.rand()

        if prob <= self.prob_crossover:

            index_prob = np.random.rand(len(parent_1.sequence))
            child_1, child_2 = deepcopy(parent_1), deepcopy(parent_2)
            for i in range(len(index_prob)):
                if index_prob[i] <= 0.5:
                    child_1.sequence[i], child_2.sequence[i] = child_2.sequence[i], child_1.sequence[i]

            return child_1, child_2

        else:
            return deepcopy(parent_1), deepcopy(parent_2)









import numpy as np


class FeatureSetter:
    """
    Chromosome class, that contains the sequence, the hyperparameters and the fitness of its linked program run
    Attributes
    ----------
    algo_type: str
               can be "hyperparameter_tuning" or "feature_selection", it represents what is the purpose of the G
    feature_num: int
                 the number of features from which to select

    Methods
    -------
    self.set_feature_sequence(): it generates a sequence of 0s and 1s according to if we want to keep or not the column
    """
    def __init__(self, algo_type: str, feature_num: int) -> None:
        self.algo_type = algo_type
        self.feature_num = feature_num

    def set_feature_sequence(self) -> list:
        """
        it generates a sequence of 0s and 1s according to if we want to keep or not the column
        Returns
        -------
        list
        a sequence of 0s and 1s with the size equal to the n_features
        """
        if self.algo_type == "feature_selection":
            sequence = np.random.choice([0, 1], size=self.feature_num, p=[0.5, 0.5])
            return sequence

        else:
            raise ValueError("If the algo_type is not feature_selection, you should not need this method")



import pandas as pd
import numpy as np
from auto_genetic.population_initializer.chromosomes import Chromosome


class FeatureSelector:
    """
    Class used to select the features
    Attributes
    ----------
    data: pd.DataFrame
          the data that we want to apply the feature selector on, it needs to have the column target
    target: str
            the name of the column that is used as target for the evaluation

    Methods
    -------
    self.feature_select(chromosome): it selects the features of the data given the sequence of the chromosome
    """
    def __init__(self, data: pd.DataFrame = None, target: str = None) -> None:
        self.data = data
        self.target = target
        self.x = data.drop(self.target, axis=1)
        self.y = data[self.target]

    def feature_select(self, chromosome: Chromosome) -> pd.DataFrame:
        """
        it selects the features of the data given the sequence of the chromosome
        Parameters
        ----------
        chromosome: Chromosome
                    a chromosome class that contains the sequence with the columns that we want to have active or not

        Returns
        -------
        pd.DataFrame
        a dataframe that contains the target column and the columns selected from the chromosome

        """
        sequence = chromosome.sequence
        filterer = np.argwhere(sequence == 1)
        x_filtered = self.x.iloc[:, filterer.flatten()]
        df_filtered = pd.concat([x_filtered, self.y], axis=1)

        chromosome.features = list(df_filtered.columns)

        return df_filtered



from abc import ABC, abstractmethod

import pandas as pd

from auto_genetic.evaluation.metrics import *


class FitnessFunctionBase(ABC):
    """
    Base Class for fitness functions
    Attributes
    ----------
    program: class
            the user defined program that has the run() and predict() methods
    target: str
            the name of the column that is used as target for the evaluation
    train_data: pd.DataFrame
                a pandas dataframe that contains the training data, including the target
    test_data: pd.DataFrame
               a pandas dataframe that contains the training data, possibly also the target (depends on the
               fitness function)

    Methods
    -------
    self.calculate_fitness(): abstract method
    """
    def __init__(self, program, train_data: pd.DataFrame, test_data: pd.DataFrame, target: str):
        self.train_data = train_data
        self.test_data = test_data
        self.program = program
        self.target = target

    @abstractmethod
    def calculate_fitness(self):
        pass


class RMSEFitness(FitnessFunctionBase):
    """
    Class to calculate Root Mean Squared Error fitness functions
    Attributes
    ----------
    program: class
            the user defined program that has the run() and predict() methods
    target: str
            the name of the column that is used as target for the evaluation
    train_data: pd.DataFrame
                a pandas dataframe that contains the training data, including the target
    test_data: pd.DataFrame
               a pandas dataframe that contains the training data, possibly also the target (depends on the
               fitness function)

    Methods
    -------
    self.calculate_fitness(): calculates the rmse of the program and then put it negative because we are solving
                              a maximisation problem
    """
    def __init__(self, program, train_data: pd.DataFrame, test_data: pd.DataFrame, target: str) -> None:
        super(RMSEFitness, self).__init__(program, train_data, test_data, target)

    def calculate_fitness(self) -> float:
        """
        runs the program and predicts the output, then it computes the RMSE and put it negative
        Returns
        -------
        float
        the negative RMSE of the program, it's negative because we always solve a maximisation problem

        """
        self.program.run(self.train_data)
        predictions = self.program.predict(self.test_data)

        error = self.test_data[self.target] - predictions

        rmse = root_mean_squared_error(error)

        return - rmse


class MAEFitness(FitnessFunctionBase):
    """
    Class to calculate Mean Absolute Error fitness functions
    Attributes
    ----------
    program: class
            the user defined program that has the run() and predict() methods
    target: str
            the name of the column that is used as target for the evaluation
    train_data: pd.DataFrame
                a pandas dataframe that contains the training data, including the target
    test_data: pd.DataFrame
               a pandas dataframe that contains the training data, possibly also the target (depends on the
               fitness function)

    Methods
    -------
    self.calculate_fitness(): calculates the mae of the program and then put it negative because we are solving
                              a maximisation problem
    """
    def __init__(self, program, train_data: pd.DataFrame, test_data: pd.DataFrame, target: str) -> None:
        super(MAEFitness, self).__init__(program, train_data, test_data, target)

    def calculate_fitness(self) -> float:
        """
        runs the program and predicts the output, then it computes the MAE and put it negative
        Returns
        -------
        float
        the negative MAE of the program, it's negative because we always solve a maximisation problem

        """
        self.program.run(self.train_data)
        predictions = self.program.predict(self.test_data)

        error = self.test_data[self.target] - predictions

        mae = root_mean_squared_error(error)

        return - mae


class SMAPEFitness(FitnessFunctionBase):
    """
    Class to calculate Scaled Mean Absolute Percentage Error fitness functions
    Attributes
    ----------
    program: class
            the user defined program that has the run() and predict() methods
    target: str
            the name of the column that is used as target for the evaluation
    train_data: pd.DataFrame
                a pandas dataframe that contains the training data, including the target
    test_data: pd.DataFrame
               a pandas dataframe that contains the training data, possibly also the target (depends on the
               fitness function)

    Methods
    -------
    self.calculate_fitness(): calculates the SMAPE of the program and then put it negative because we are solving
                              a maximisation problem
    """
    def __init__(self, program, train_data: pd.DataFrame, test_data: pd.DataFrame, target: str) -> None:
        super(SMAPEFitness, self).__init__(program, train_data, test_data, target)

    def calculate_fitness(self) -> float:
        """
        runs the program and predicts the output, then it computes the SMAPE and put it negative
        Returns
        -------
        float
        the negative SMAPE of the program, it's negative because we always solve a maximisation problem

        """
        self.program.run(self.train_data)
        predictions = self.program.predict(self.test_data)

        s_mape = symmetric_mean_average_percentage_error(self.test_data[self.target], predictions)

        return - s_mape


class MAAPEFitness(FitnessFunctionBase):
    """
    Class to calculate Mean Arctan Absolute Percentage Error fitness functions
    Attributes
    ----------
    program: class
            the user defined program that has the run() and predict() methods
    target: str
            the name of the column that is used as target for the evaluation
    train_data: pd.DataFrame
                a pandas dataframe that contains the training data, including the target
    test_data: pd.DataFrame
               a pandas dataframe that contains the training data, possibly also the target (depends on the
               fitness function)

    Methods
    -------
    self.calculate_fitness(): calculates the MAAPE of the program and then put it negative because we are solving
                              a maximisation problem
    """

    def __init__(self, program, train_data: pd.DataFrame, test_data: pd.DataFrame, target: str) -> None:
        super(MAAPEFitness, self).__init__(program, train_data, test_data, target)

    def calculate_fitness(self) -> float:
        """
        runs the program and predicts the output, then it computes the MAAPE and put it negative
        Returns
        -------
        float
        the negative MAAPE of the program, it's negative because we always solve a maximisation problem

        """
        self.program.run(self.train_data)
        predictions = self.program.predict(self.test_data)

        s_mape = mean_arctangent_absolute_percentage_error(self.test_data[self.target], predictions)

        return - s_mape



import numpy as np
import pandas as pd


def root_mean_squared_error(error: pd.Series) -> float:
    """
    Function to calculate the rmse of a pd series
    Parameters
    ----------
    error: pd.Series
           pandas series that contains the errors for the model

    Returns
    -------
    float
    the rmse for the specific chromosome we are analysing

    """
    return np.sqrt((error ** 2).mean())


def mean_absolute_error(error: pd.Series) -> float:
    """
    Function to calculate the mae of a pd series
    Parameters
    ----------
    error: pd.Series
          pandas series that contains the errors for the model

    Returns
    -------
    float
    the mae for the specific chromosome we are analysing

    """
    return abs(error).mean()


def symmetric_mean_average_percentage_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    """
    Function to calculate the smape of a pd series
    Parameters
    ----------
    y_true: pd.Series
            pandas series that contains the true values for the test set
    y_pred: pd.Series
            pandas series that contains the predicted values for the test set

    Returns
    -------
    float
    the smape for the specific chromosome we are analysing

    """
    errors = 2 * np.abs(y_pred - y_true) / (np.abs(y_true) + np.abs(y_pred)) * 100
    return errors.mean()


def mean_arctangent_absolute_percentage_error(y_true: pd.Series, y_pred: pd.Series) -> float:
    """
    Function to calculate the maape of a pd series
    Parameters
    ----------
    y_true: pd.Series
            pandas series that contains the true values for the test set
    y_pred: pd.Series
            pandas series that contains the predicted values for the test set

    Returns
    -------
    float
    the maape for the specific chromosome we are analysing

    """
    aape = np.arctan(abs((y_true - y_pred) / y_true))
    return aape.mean()



from __future__ import annotations
import pandas as pd

from auto_genetic.evaluation.fitness_functions import *


class FitnessFuncSetter:
    """
    Class that sets the fitness function for the chromosomes
    Attributes
    ----------
    program: class
            the user defined program that has the run() and predict() methods
    evaluation_method: str
                       the string that defines the evaluation method that we want to use inside the evaluator,
                       if it's 'custom' the user needs to define the custom_fitness_function
    target: str
            the name of the column that is used as target for the evaluation
    train_data: pd.DataFrame
                a pandas dataframe that contains the training data, including the target
    test_data: pd.DataFrame
               a pandas dataframe that contains the training data, possibly also the target (depends on the
               fitness function)

    Methods
    -------
    self.set_fitness_func(custom_fitness_func): method that allows the user to pass a custom fitness function
    self.get_fitness_func(): method that sets a fitness function for a chromosome
    """
    def __init__(self, evaluation_method: str, program, train_data: pd.DataFrame, test_data: pd.DataFrame, target: str)\
            -> None:
        self.evaluation_method = evaluation_method
        self.program = program
        self.train_data = train_data
        self.test_data = test_data
        self.target = target
        self.fitness_func = None

    def set_fitness_func(self, custom_fitness_function=None) -> FitnessFuncSetter:
        """
        method that allows the user to pass a custom fitness function
        Parameters
        ----------
        custom_fitness_function: class
                                 user defined fitness function, it can be used if the evaluation_method is "custom"

        Returns
        -------
        FitnessFuncSetter
        it returns the class itself, so it can be used inline

        """
        if self.evaluation_method == "custom":
            if custom_fitness_function is None:
                raise ValueError("to use the custom method you need to implement a custom fitness function"
                                 "inheriting the FitnessFunction class or (at your own risk) by implementing it freely")

            self.fitness_func = custom_fitness_function(self.program, self.train_data, self.test_data, self.target)

        elif self.evaluation_method == "rmse":
            self.fitness_func = RMSEFitness(self.program, self.train_data, self.test_data, self.target)

        elif self.evaluation_method == "mae":
            self.fitness_func = MAEFitness(self.program, self.train_data, self.test_data, self.target)

        elif self.evaluation_method == "s_mape":
            self.fitness_func = SMAPEFitness(self.program, self.train_data, self.test_data, self.target)

        elif self.evaluation_method == "maape":
            self.fitness_func = MAAPEFitness(self.program, self.train_data, self.test_data, self.target)

        else:
            raise ValueError("The selected evaluation method is not available, choose one of the available ones or "
                             "implement your own by choosing custom")

        return self

    def get_fitness_func(self):
        """
        method that allows the user to pass a custom fitness function
        Returns
        -------
        class
        the initialised fitness function that it's chosen
        """
        if self.fitness_func:
            return self.fitness_func

        else:
            raise ValueError("The fitness function has not been defined")



import numpy as np
from auto_genetic.population_initializer.chromosomes import Chromosome
from copy import deepcopy


class Translation:
    """
    Class that performs the translation over a Chromosome
    Attributes
    ----------
    prob_translation: float
                      the probability that the translation is performed
    Methods
    -------
    self.perform_translation(child): it uses the Translation class on a Chromosome and performs the translation
    """
    def __init__(self, prob_translation: float) -> None:
        self.prob_translation = prob_translation

    def perform_translation(self, child: Chromosome) -> Chromosome:
        prob = np.random.rand()
        translated_gene = deepcopy(child)
        if prob <= self.prob_translation:
            num_traslated = np.random.randint(0, len(child.sequence))
            translated_gene.sequence = np.hstack([translated_gene.sequence[-num_traslated:],
                                                  translated_gene.sequence[:-num_traslated]])

        translated_gene.fitness = None

        return translated_gene



import pandas as pd
from abc import ABC, abstractmethod


class BaseProgram(ABC):
    """
    Abstract Base Class for the programs, any program has to inherit from this one, or have the same methods, is
    possible to add any additional method, but they need to be run inside either run or predict
    Methods
    -------
    self.set_program_hyperparams(hyperparams): method that is used by each chromosome to set the hyperparameter for
                                               the program, os it needs to be there
    self.run(train_data): it runs the defined program, all the processes of the program have to be passed here
    self.predict(test_data): it predicts and these predictions are used to calculate the fitness of the program
    """
    def __init__(self):
        self.hyperparams = None

    def set_program_hyperparams(self, hyperparams: dict):
        self.hyperparams = hyperparams

    @abstractmethod
    def run(self, train_data: pd.DataFrame):
        pass

    @abstractmethod
    def predict(self, test_data: pd.DataFrame):
        pass



import numpy as np
import random
import warnings


class Selection:
    """
        Class that performs the selection over the whole population
        Attributes
        ----------
        population: list
                    the list that contains all the Chromosomes that make up the population
        selection_method: str
                          the method that we can use to perform the selection of the parents, it can be "roulette_wheel",
                          "stochastic_universal_sampling", "tournament", "rank" or "random"
        reproduction_rate: flaat
                           the rate at which the population reproduces itself, the number of children is equal to number
                           of chromosome in the generation * reproduction_rate
        Methods
        -------
        Private:
        self._roulette_wheel_selection(): it performs the roulette wheel selection on the population to get the parents
        self._stochastic_universal_sampling(): it performs the stochastic universal sampling selection on the
                                               population to get the parents
        self._tournament_selection(tournament_size): it performs the tournament selection on the population
                                                     to get the parents
        self._rank_selection(): it performs the rank selection on the population to get the parents
        self._random_selection(): it performs the random selection on the population to get the parents
        self._get_subset_sum(gene_num): it adds up the total fitness of the population up to the gene_num specified
        Public:
            self.perform_selection(tournament_size): it performs the selection by feeding the parents to the
                                                    previously specified selection method
    """

    def __init__(self, selection_method: str, population: list, reproduction_rate: float) -> None:
        self.selection_method = selection_method
        self.population = population
        self.reproduction_rate = reproduction_rate
        self.n_parents: int = int(len(self.population) * self.reproduction_rate)

    def perform_selection(self, tournament_size: int = None) -> np.array:
        """
        it performs the selection by feeding the parents to the previously specified selection method
        Parameters
        ----------
        tournament_size: int
                         the size of the tournament that we use for the "tournament" selection
        Returns
        -------
        np.array
        an array that contains the selected parents for the reproduction
        """
        if self.selection_method == "roulette_wheel":
            parents = self._roulette_wheel_selection()
        elif self.selection_method == "stochastic_universal_sampling":
            parents = self._stochastic_universal_sampling()
        elif self.selection_method == "tournament":
            if tournament_size is None:
                raise ValueError("We need a valid tournament_size for this method")
            parents = self._tournament_selection(tournament_size=tournament_size)
        elif self.selection_method == "rank":
            parents = self._rank_selection()
        elif self.selection_method == "random":
            parents = self._random_selection()
        else:
            raise ValueError("The selection method specified is not implemented \n"
                             "Please specify a valid selection method ")

        return parents

    def _roulette_wheel_selection(self) -> np.array:
        """
        it performs the roulette wheel selection on the population to get the parents
        Returns
        -------
        np.array
        an array that contains the selected parents for the reproduction
        """
        all_parents = []
        while len(all_parents) < self.n_parents:

            population_fitness = sum([chromosome.fitness for chromosome in self.population])

            chromosome_probabilities = [chromosome.fitness / population_fitness for chromosome in self.population]

            all_parents.append(np.random.choice(self.population, p=chromosome_probabilities))

        return np.array(all_parents)

    def _stochastic_universal_sampling(self) -> np.array:
        """
        it performs the stochastic universal sampling selection on the population to get the parents
        Returns
        -------
        np.array
        an array that contains the selected parents for the reproduction
        """
        total_fitness = sum([chromosome.fitness for chromosome in self.population])
        point_distance = total_fitness / self.n_parents
        start_point = random.uniform(0, point_distance)
        points = [start_point + i * point_distance for i in range(self.n_parents)]

        parents = set()
        while len(parents) < self.n_parents:
            random.shuffle(self.population)
            n_rounds = 0
            while n_rounds < len(points) and len(parents) < self.n_parents:
                for gene_num in range(len(self.population)):
                    if self._get_subset_sum(gene_num) > points[n_rounds]:
                        parents.add(self.population[gene_num])
                        break
                n_rounds += 1

        return np.array(parents)

    def _get_subset_sum(self, gene_num: int) -> float:
        """
        it adds up the total fitness of the population up to the gene_num specified
        Parameters
        ----------
        gene_num: the position up to which to sum the fitness for the population

        Returns
        -------
        float
        the total fitness for the population up to gene num
        """
        subset_sum = 0.0
        i = 0

        while i <= gene_num:
            subset_sum += self.population[i].fitness
            i += 1
        return subset_sum

    def _tournament_selection(self, tournament_size=4):
        """
        it performs the tournament selection on the population to get the parents
        Parameters
        ----------
        tournament_size: int
                         the size of the tournament that we use for the tournament selection

        Returns
        -------
        np.array
        an array that contains the selected parents for the reproduction
        """
        all_parents = []
        while len(all_parents) < self.n_parents:
            parents = random.choices(self.population, k=tournament_size)
            parents = sorted(parents, key=lambda gene: gene.fitness, reverse=True)
            all_parents.append(parents[0])

        return np.array(all_parents)

    def _rank_selection(self):
        """
        it performs the rank selection on the population to get the parents
        Returns
        -------
        np.array
        an array that contains the selected parents for the reproduction
        """
        parents = sorted(self.population, key=lambda gene: gene.fitness, reverse=True)
        best_parents = parents[:self.n_parents]

        random.shuffle(best_parents)

        return np.array(best_parents)

    def _random_selection(self):
        """
        it performs the random selection on the population to get the parents
        Returns
        -------
        np.array
        an array that contains the selected parents for the reproduction
        """
        warnings.warn("The random selections leads to a stochastic search for the parameters")
        parents = random.choices(self.population, k=self.n_parents)

        return np.array(parents)



from __future__ import annotations

from auto_genetic.population_initializer.hyperparams_setter import HyperParamsSetter
from auto_genetic.population_initializer.feature_setter import FeatureSetter
from auto_genetic.evaluation.fitnessfuncsetter import FitnessFuncSetter
from typing import List


class Chromosome:
    """
    Chromosome class, that contains the sequence, the hyperparameters and the fitness of its linked program run
    Attributes
    ----------
    algo_type: str
               can be "hyperparameter_tuning" or "feature_selection", it represents what is the purpose of the GA
    hyperparams_dict: List[dict, dict]
                      a list that contains 2 dictionaries, one with the hyperparameters' types and another with the
                      hyperparameters' values
    feature_num: int
                 the number of features from which to select
    sequence: list
              the list of the values of the chromosome's gene
    hyperparams: dict
                 the dictionary that contains the hyperparameters linked to this specific Chromosome
    fitness: float
             the fitness of the chromosome


    Methods
    -------
    Private:
        self._initialize_hyperparams(): it sets the sequence of the chromosome given the set of hyperparameters
        self._initialize_features(): it sets the sequence of the chromosome given the number of features
    Public:
        self.initialize(): method that initializes the sequence of the chromosome according to the algo_type
        self.set_fitness_function(fitness_function_initialised): method that sets the fitness function for the
                                                                 chromosome
        self.calculate_fitness(): method that uses the previously set fitness function to calculate the fitness for
                                  this specific chromosome
    """
    def __init__(self, algo_type: str, hyperparams_dict: List[dict, dict] = None, feature_num: int = None) -> None:
        self.fitness = None
        self.fitness_function = None
        self.algo_type = algo_type
        if self.algo_type == "hyperparameter_tuning" and hyperparams_dict is None:
            raise ValueError("We don't have any hyerparameters to tune!!")

        elif self.algo_type == "feature_selection" and feature_num is None:
            raise ValueError("What's the number of features you want to tune?")

        self.hyperparams_dict = hyperparams_dict
        self.feature_num = feature_num
        self.hyperparams = None
        self.sequence = None
        self.hyperparams_map = None
        self.features = None

    def initialize(self) -> Chromosome:
        """
        method that initializes the sequence of the chromosome according to the algo_type
        Returns
        -------
        Chromosome
        it returns the class itself, so it can be used inline
        """
        if self.algo_type == "hyperparameter_tuning":
            self._initialize_hyperparams()
        elif self.algo_type == "feature_selection":
            self._initialize_feature()

        return self

    def set_fitness_function(self, fitness_function_initialised: FitnessFuncSetter) -> None:
        """
        method that sets the fitness function for the chromosome
        Parameters
        ----------
        fitness_function_initialised: FitnessFuncSetter
                                      The initialised fitness function that contains the program to run

        """
        self.fitness_function = fitness_function_initialised

    def calculate_fitness(self):
        """
        method that uses the previously set fitness function to calculate the fitness for this specific chromosome
        """
        if self.fitness:
            return self.fitness
        else:
            self.fitness = self.fitness_function.calculate_fitness()

    def _initialize_hyperparams(self) -> None:
        """
        it sets the sequence of the chromosome given the set of hyperparameters
        """
        setter = HyperParamsSetter(self.hyperparams_dict, self.algo_type).get_hyperparameters()
        self.sequence = setter.convert_hyperparams_values()
        self.hyperparams, self.hyperparams_map = setter.get_program_hyperparams()

    def _initialize_feature(self) -> None:
        """
        it sets the sequence of the chromosome given the number of features
        """
        setter = FeatureSetter(self.algo_type, self.feature_num)
        self.sequence = setter.set_feature_sequence()



from typing import Iterable

from tqdm import tqdm
import pandas as pd

from auto_genetic.reproduction.reproduction import Reproduction
from auto_genetic.evaluation.evaluator import Evaluator
from auto_genetic.darwinism import Darwinism
from auto_genetic.population_initializer.chromosomes import Chromosome
from auto_genetic.population_initializer.population import PopulationInitializer
from sklearn.model_selection import train_test_split


class GeneticAlgorithm:
    """
    Class that runs the genetic algorithm for the specified program
    Attributes
    ----------
    program: class
        the user defined program that has the run() and predict() methods
    data: pd.DataFrame
          a pandas dataframe that contains the data for the program, including the target
    target_column: str
                   the name of the column that is used as target for the evaluation
    test_size: float
              the size of the test data for the testing of the chromosomes
    data_split: str
                it chooses if we want to split the data one time for all the generations or for each generation,
                it can take the values "single" or "multiple"
    algo_type: str
               can be "hyperparameter_tuning" or "feature_selection", it represents what is the purpose of the GA
    pop_size: int
              the number of chromosomes we want to have in each generation
    number_gen: int
                the number of generations we want the program to run for
    hyperparams_dict: List[dict, dict]
                      a list that contains 2 dictionaries, one with the hyperparameters' types and another with the
                      hyperparameters' values
    feature_num: int
                the number of features that the feature_selection algorithm can select from
    max_fitness_value: float
                       the maximum value that we want our fitness to take, if it gets higher the algorithm stops
    prob_crossover: float
                    the probability that the crossover is performed, by default it's equal to 1, so the crossover
                    will always be applied
    crossover_method: str
                      the methods that we can use to perform the crossover, it can be "single_point_split",
                      "two_point_split" or "uniform"
    mutation_method: str
                      the methods that we can use to perform the mutation, it can be "bit_flip", "random",
                      "swap", "scramble" or "inversion"
    prob_mutation: float
                    the probability that the mutation is performed
    prob_translation: float
                      the probability that the translation is performed
   reproduction_rate: flaat
                       the rate at which the population reproduces itself, the number of children is equal to number
                       of chromosome in the generation * reproduction_rate
    tournament_size: int
                     the size of the tournament that we use for the "tournament" selection
    selection_method: str
                      the method that we can use to perform the selection of the parents, it can be
                      "roulette_wheel", "stochastic_universal_sampling", "tournament", "rank" or "random"

    Methods
    -------
    self.initialize_population(): method that uses the Chromosome class to generate n (= pop_size) Chromosomes,
                                  passing by the PopulationInitializer class
    self.split_data(): Method that defines the train and test datasets
    self.set_evaluation_method(evaluation_method, custom_fitness_function): Method that sets the fitness function
                                                                            for the program
    self.evaluate_generation(): Method to evaluate the full generation by finding the fitness for every member of
                                the generation
    self.reproduction(): it puts together all the steps necessary for the reproduction and generates the children
                         that will be added to the population
    self.darwinism(): it eliminates the chromosomes with the worst fitness from the generation, to make space
                      for the children
    self.run(): run the genetic algorithm for n generation with m genes, storing the best score and the best gene
    """
    def __init__(self, program, data: pd.DataFrame, target_column: str, test_size: float = 0.2,
                 data_split: str = "single", algo_type: str = "hyperparameter_tuning", pop_size: int = 100,
                 number_gen: int = 20, hyperparams_dict: list[dict, dict] = None, feature_num: int = None,
                 max_fitness_value: float = None,
                 prob_crossover: float = 1.0, crossover_method: str = "single_point_split",
                 mutation_method: str = "bit_flip", prob_mutation: float = 0.3, prob_translation: float = 0.1,
                 reproduction_rate: float = 0.2, selection_method: str = "roulette_wheel", tournament_size: int = 4):

        self.algo_type = algo_type
        self.pop_size = pop_size
        self.number_gen = number_gen
        self.max_fitness_value = max_fitness_value
        self.prob_crossover = prob_crossover
        self.crossover_method = crossover_method
        self.prob_mutation = prob_mutation
        self.prob_translation = prob_translation
        self.reproduction_rate = reproduction_rate
        self.mutation_method = mutation_method
        self.hyperparams_dict = hyperparams_dict
        if self.hyperparams_dict:
            self.hyperparams_names = self.hyperparams_dict[0].keys()
            self.hyperparams_values = self.hyperparams_dict[1]
            self.hyperparams_types = self.hyperparams_dict[0]
        else:
            self.hyperparams_names = None
            self.hyperparams_values = None
            self.hyperparams_types = None
        self.population = None
        self.selection_method = selection_method
        self.best_chromosome = None
        self.tournament_size = tournament_size
        self.evaluation_method = None
        self.program = program
        self.target = target_column
        self.train_data = None
        self.test_data = None
        self.custom_fitness_function = None
        self.feature_num = feature_num
        self.data = data
        self.test_size = test_size
        self.data_split = data_split

    def initialize_population(self) -> None:
        """
        method that uses the Chromosome class to generate n (= pop_size) Chromosomes, passing by the
        PopulationInitializer class
        """
        self.population = PopulationInitializer(self.pop_size, self.algo_type, hyperparams_dict=self.hyperparams_dict,
                                                feature_num=self.feature_num).initialize_population()

    def split_data(self) -> None:
        """
        Method that defines the train and test datasets
        """
        if self.data_split == "single":
            if self.train_data is None and self.test_data is None:
                self.train_data, self.test_data = train_test_split(self.data, test_size=self.test_size)
        elif self.data_split == "multiple":
            self.train_data, self.test_data = train_test_split(self.data, test_size=self.test_size)

        else:
            raise ValueError("The type of data split specified does not exist")

    def set_evaluation_method(self, evaluation_method: str, custom_fitness_function=None) -> None:
        """
        Method that sets the fitness function for the program
        Parameters
        ----------
        evaluation_method: str
                           the string that defines the evaluation method that we want to use inside the evaluator,
                           if it's 'custom' the user needs to define the custom_fitness_function
        custom_fitness_function: class
                                 if evaluation_method is custom here we have to have a custom evaluation function
        """

        if self.evaluation_method == "custom" and custom_fitness_function is None:
            raise ValueError("To use a custom evaluation method we need a custom fitness function")

        self.evaluation_method = evaluation_method
        self.custom_fitness_function = custom_fitness_function

    def evaluate_generation(self) -> None:
        """
        Method to evaluate the full generation by finding the fitness for every member of the generation
        """
        evaluator = Evaluator(self.program, self.population, self.evaluation_method, self.target, self.algo_type,
                              self.train_data, self.test_data, self.custom_fitness_function)

        evaluator.evaluate_generation()

    def reproduction(self) -> Iterable:
        """
        it puts together all the steps necessary for the reproduction and generates the children that
        will be added to the population
        Returns
        -------
        Iterable
        a list or numpy array that contains the new chromosomes to be added to the next generation
        """

        rep_function = Reproduction(population=self.population, crossover_method=self.crossover_method,
                                    prob_crossover=self.prob_crossover, hyperparams_types=self.hyperparams_types,
                                    mutation_method=self.mutation_method,
                                    prob_mutation=self.prob_mutation, prob_translation=self.prob_translation,
                                    selection_method=self.selection_method, reproduction_rate=self.reproduction_rate,
                                    tournament_size=self.tournament_size)
        children = rep_function.reproduction()

        return children

    def darwinism(self) -> None:
        """
        it eliminates the chromosomes with the worst fitness from the generation, to make space for the children
        """

        pop_selector = Darwinism(self.population, self.reproduction_rate)
        self.population, self.best_chromosome = pop_selector.discard_worst_performers()

    def run(self) -> Chromosome:
        """
        runs the genetic algorithm for n generation with m genes, storing the best score and the best gene
        """
        self.initialize_population()

        if self.evaluation_method is None:
            raise ValueError("Remember to set the evaluation method before running the algorithm")

        for _ in tqdm(range(self.number_gen)):

            self.split_data()
            self.evaluate_generation()
            children = self.reproduction()

            self.darwinism()

            self.population = self.population + children

            if self.max_fitness_value:
                if self.best_chromosome.fitness > self.max_fitness_value:
                    break

        return self.best_chromosome
